''' dAmnViper.data module
    This module is part of the dAmnViper package.
    Created by photofroggy.
'''

from dAmnViper.parse import Packet

class Channel:
    """ Objects representing dAmn channels.
        Information about the channels are stored in here.
    """
    
    class header:
        def __init__(self):
            self.content = ''
            self.by = ''
            self.ts = 0.0
    
    def __init__(self, namespace, shorthand):
        """Set up all our variables."""
        self.title = Channel.header()
        self.topic = Channel.header()
        self.pc = {}
        self.pc_order = []
        self.member = {}
        
        self.namespace = namespace
        self.shorthand = shorthand
        
        self.type = '<dAmn channel \''+self.namespace+'\'>'
    
    def process_property(self, data):
        if data['p'] == 'title':
            self.title.content = data['value']
            self.title.by = data['by']
            self.title.ts = data['ts']
        if data['p'] == 'topic':
            self.topic.content = data['value']
            self.topic.by = data['by']
            self.topic.ts = data['ts']
        if data['p'] == 'privclasses':
            self.pc = Packet(data['value'], ':').args
            self.pc_order = sorted(self.pc.keys(), key=int)
            self.pc_order.reverse()
        if data['p'] == 'members':
            member = Packet(data['value'])
            while member.cmd != None and len(member.args) > 0:
                self.register_user(member)
                member = Packet(member.body)
    
    def register_user(self, info, user = None):
        user = user if user != None else info.param
        if user in self.member:
            self.member[user]['con']+= 1
        else:
            self.member[user] = info.args
            self.member[user]['con'] = 1
    
    def __str__(self):
        return self.namespace

# EOF


''' dAmnViper.base module
    This module is part of the dAmnViper package.
    Created by photofroggy.
'''

import sys
import time
# Import dependencies from the API!
from dAmnViper.data import Channel
from dAmnViper.parse import Packet
from dAmnViper.parse import Protocol
from dAmnViper.stream import Connection
from dAmnViper.deviantART import Login

class dAmnSock:
    """ dAmnSock class.
        This is a base class for managing a connection
        to a dAmn server.
    """
    
    class platform:
        """Information about the dAmn Viper platform."""
        name = 'dAmn Viper'
        version = 2
        state = 'RC'
        build = 54
        stamp = '29052011-122308'
        series = 'Swift'
        author = 'photofroggy'
            
    class user:
        """User login data."""
        username = None
        password = None
        cookie = None
        token = None

    class flag:
        """Status flags are stored here"""
        connecting = False
        shaking = False
        loggingin = False
        connected = False
    
    class CONST:
        """dAmnSock style 'constants'."""
        SERVER = 'chat.deviantart.com'
        CLIENT = 'dAmnClient 0.3'
        PORT = 3900
    
    class session:
        status = False
        cookie = None
        token = None

    extras = {'remember_me':'1'}
    agent = 'dAmnViper (python 3.x) dAmnSock/1.1'
    sock = None
    
    def nullflags(self):
        self.flag.connecting = False
        self.flag.shaking = False
        self.flag.loggingin = False
        self.flag.connected = False
    
    def logger(self, ns, msg, showns=True, mute=False, pkt=None):
        if mute: return
        stamp = time.strftime('%H:%M:%S|')
        sys.stdout.write('>> {0}{1} {2}\n'.format(stamp, (ns + '|' if showns else ''), msg))
        sys.stdout.flush()
    
    def get_token(self):
        self.on_get_token()
        if not self.user.password:
            return
        self.session = Login(self.user.username, self.user.password, self.extras, self.agent)
        if self.session.status[0] == 1:
            self.user.cookie = self.session.cookie
            self.user.token = self.session.token
            self.user.password = None
        self.on_token()

    def on_get_token(self):
        """Overwrite this method to do shit when get_token is called."""
        pass
    
    def on_token(self):
        """Overwrite this method to do stuff after get_token is finished."""
        pass
    
    def connect(self):
        """Open a connection to dAmn!"""
        if self.sock and self.sock.sock:
            return
        self.sock = Connection(self.CONST.SERVER, self.CONST.CLIENT, self.CONST.PORT)
        self.sock.connect()
        self.flag.connecting = self.sock.connected()
        return self.flag.connecting
    
    def handshake(self, vals=None):
        # Should be able to define values sent in the handshake packet.
        if not self.flag.connecting:
            return False
        self.flag.shaking = True # Raise the flag, hoe.
        shake = self.CONST.CLIENT
        shake+= '\nagent='+self.agent
        if vals == None:
            shake+= '\n'
        else:
            shake+= "\n{0}\n".format("\n".join(['='.join([key, vals[key]]) for key in vals]))
        if len(shake+'\0') == self.send(shake):
            return True
        return False
    
    def login(self):
        # Send our login packet. Set the loggingin flag to true.
        self.flag.loggingin = True
        return bool(self.send('login {0}\npk={1}\n'.format(self.user.username, self.user.token)))
    
    def send(self, data):
        # Send data to dAmn!
        if self.sock == None or not self.sock.connected():
            return 0
        return self.sock.send(data)
    
    def get_packet(self):
        # Return a single packet from the queue.
        if not self.sock.connected():
            self.nullflags()
            return None
        return self.sock.get_packet()
    
    def close(self):
        # This is how we close our connection!
        if self.sock and self.sock.connected():
            self.sock.close()
    
    # PROTOCOL!
    # The methods below pretty much define the protocol for outgoing packets.

    def raw(self, data):
        # Send a raw dAmn packet.
        return self.send(data)
        
    def pong(self):
        # dAmn likes to play ping pong...
        return self.send('pong\n')
        
    def join(self, ns):
        # Send a join packet to dAmn.
        return self.send('join {0}\n'.format(ns))
        
    def part(self, ns):
        # Send a part packet to dAmn
        return self.send('part {0}\n'.format(ns))
        
    def say(self, ns, message):
        # Send a message to a dAmn channel namespace.
        return self.send('send {0}\n\nmsg main\n\n{1}'.format(ns, str(message)))
    
    def action(self, ns, message):
        # Send an action to a dAmn channel namespace.
        return self.send('send {0}\n\naction main\n\n{1}'.format(ns, str(message)))
    
    def me(self, ns, message):
        # This is just another way to do an action.
        return self.action(ns, message)
        
    def npmsg(self, ns, message):
        # Send a non-parsed message to a dAmn channel namespace.
        return self.send('send {0}\n\nnpmsg main\n\n{1}'.format(ns, str(message)))
    
    def promote(self, ns, user, pc=None):
        # Promote a user in a dAmn channel.
        return self.send('send {0}\n\npromote {1}\n{2}'.format(ns, user, '' if pc == None else '\n'+str(pc)))
    
    def demote(self, ns, user, pc=None):
        # Demote a user in a dAmn channel.
        return self.send('send {0}\n\ndemote {1}\n{2}'.format(ns, user, '' if pc == None else '\n'+str(pc)))
    
    def kick(self, ns, user, r=None):
        # Kick a user out of a dAmn channel.
        return self.send('kick {0}\nu={1}\n{2}'.format(ns, user, '' if r == None else '\n'+str(r)))
    
    def ban(self, ns, user):
        # Ban a user from a dAmn channel.
        return self.send('send {0}\n\nban {1}\n'.format(ns, user))
    
    def unban(self, ns, user):
        # Unban someone from a dAmn channel.
        return self.send('send {0}\n\nunban {1}\n'.format(ns, user))
    
    def get(self, ns, p):
        # Get a property for a dAmn channel.
        return self.send('get {0}\np={1}\n'.format(ns, p))
    
    def set(self, ns, p, val):
        # Set a property for a dAmn channel.
        return self.send('set {0}\np={1}\n\n{2}'.format(ns, p, val))
    
    def admin(self, ns, command):
        # Send an admin command to a dAmn channel. No need for multiple methods here.
        return self.send('send {0}\n\nadmin\n\n{1}'.format(ns, command))
    
    def disconnect(self):
        # Send a disconnect packet to the dAmn server.
        return self.send('disconnect\n')
        
    def kill(self, ns, r=None):
        # Send a kill packet to the dAmn server.
        return self.send('kill {0}\n{1}'.format(ns, '' if r == None else '\n'+str(r)))
    
    # END PROTOCOL METHODS
    
    def parser(self):
        # Just a shortcut for getting the Packet class.
        return Packet

class Client(dAmnSock):
    """ dAmn Client class.
        This provides a more client-like interface between
        the dAmn server and the application.
    """

    agent = 'dAmnViper (Python 3.x) Client/dAmnSock/x.x'
    channel = {}
    _protocol = Protocol

    def __init__(self, *args, **kwargs):
        # Create an instance of our protocol class. Do anything else required.
        self.protocol = self._protocol()
        self.__inst__(*args, **kwargs)
        self._lastp = time.time()
    
    def __inst__(self, *args, **kwargs):
        """Overwrite this method!"""
        self.agent = 'dAmnViper (Python 3.1) Client/dAmnSock/{0}.{1}'.format(
            self.platform.version, self.platform.build)
    
    def logger(self, ns, msg, showns=True, mute=False, pkt=None):
        # Prints messages to stdout.
        if mute: return
        stamp = time.strftime('%H:%M:%S|')
        sys.stdout.write('>> {0}{1} {2}\n'.format(stamp, (ns + '|' if showns else ''), msg))
        sys.stdout.flush()
    
    def format_ns(self, ns):
        # This takes a dAmn channel name and formats it as a raw dAmn namespace.
        ns = str(ns)
        un = self.user.username
        pre = ns[:1]
        if pre == '#':
            return 'chat:{0}'.format(ns[1:])
        if pre == '@':
            if not un:
                return 'pchat:{0}'.format(ns[1:])
            para = [ns[1:], un]
            para.sort(key=str.lower)
            return 'pchat:{0}:{1}'.format(para[0], para[1])
        if ns[:5] == 'chat:' or ns[:6] == 'pchat:':
            return ns
        return 'chat:'+ns
    
    def deform_ns(self, ns):
        # This does the opposite of format_ns()
        parts = str(ns).split(':')
        discard = self.user.username
        if parts[0].lower() == 'chat':
            return '#{0}'.format(parts[1])
        if parts[0].lower() == 'pchat':
            if not discard:
                return '@{0}:{1}'.format(parts[1], parts[2])
            parts.pop(0)
            for i in range(0,2):
                if parts[i].lower() == discard.lower():
                    parts.pop(i)
                    return '@{0}'.format(parts[0])
        if ns[:1] in ('#', '@'):
            return ns
        return '#'+ns

    def get_packets(self):
        # This implementation is simply faster than get_packet().
        retd = []
        nmeth = lambda data: data
        pkts = self.sock.get_packets()
        for raw_pkt in pkts:
            retd.append(self.process_packet(raw_pkt, nmeth))
        if retd:
            self._lastp = time.time()
        else:
            if (time.time() - self._lastp) >= 120:
                self._timeout()
        return retd
    
    def process_packet(self, raw_pkt, nmeth):
        pkt = Packet(raw_pkt)
        ns = '~Global'
        if pkt.param:
            if pkt.param[:5] in ('chat:', 'pchat'):
                ns = self.deform_ns(pkt.param)
            elif pkt.param[:6] == 'login:':
                ns = '@' + pkt.param[6:]
        data = self.protocol.mapper(pkt)
        loglist = self.protocol.logger(data['event'], data['rules'], ns, pkt)
        if loglist is not None:
            self.logger(*loglist)
        getattr(self, 'pkt_' + data['event'], nmeth)(data['args'])
        self.pkt_generic(data)
        return data
    
    def run(self, *args, **kwargs):
        """Overwrite this method to provide your own application with a main loop."""
        if not self.sock:
            return
        while self.sock.connected():
            try:
                self.mainloop(*args, **kwargs)
                self.on_loop(*args, **kwargs)
            except KeyboardInterrupt as e:
                sys.stdout.write('\n')
                self.sock.close()
                self.process_packet('disconnect\ne=force quit (^C)\n\n', lambda data: data)
                break
        while len(self.sock.packet) > 0:
            self.mainloop(*args, **kwargs)
            self.on_loop(*args, **kwargs)
    
    def mainloop(self, *args, **kwargs):
        """Overwrite this to do stuff on every iteration of the main loop!"""
        self.get_packets()

    def _timeout(self):
        self.sock.packet.append('disconnect\ne=socket timed out\n\n')
    
    def on_loop(self, *args, **kwargs):
        # Overwrite this method to do stuff at the end of each main loop.
        pass
    
    # Packet handling stuffs. Only basic shizz.
    
    def pkt_generic(self, data):
        # Overwrite this method to do stuff on every packet event.
        pass
    
    def pkt_dAmnServer(self, data):
        # What to do in the event of a handshake.
        self.flag.shaking = False
        self.login()
    
    def pkt_login(self, data):
        # What to do when we receive a login packet.
        self.flag.loggingin = False
        if data['e'] == 'ok':
            self.flag.loggingin = False
            self.flag.connecting = False
            self.flag.connected = True
            return
        self.nullflags()
        self.close()
    
    def pkt_join(self, data):
        # The client tried to join a channel.
        if data['e'] == 'ok':
            self.channel[data['ns']] = Channel(
                data['ns'], self.deform_ns(data['ns'])
            )
            return
    
    def pkt_part(self, data):
        # This is what happens when we receive a part packet.
        if data['ns'] in self.channel.keys():
            del self.channel[data['ns']]
        if len(self.channel) > 0:
            return
        if 'r' in data.keys():
            if data['r'] in ('bad data', 'bad msg', 'msg too big') or 'killed:' in data['r']:
                self.sock.packet.append('disconnect{0}e={1}{2}{3}'.format("\n", data['r'], "\n", "\n"))
                return
    
    def pkt_property(self, data):
        # Make sure properties are stored in the right places.
        if data['p'] == 'info': return
        if not data['ns'] in self.channel.keys(): return
        self.channel[data['ns']].process_property(data)
    
    def pkt_recv_join(self, data):
        # When a user joins, log them!
        self.channel[data['ns']].register_user(Packet(data['info']), data['user'])
        
    def pkt_recv_part(self, data):
        # No need to log users after they leave a channel.
        if not data['user'] in self.channel[data['ns']].member: return
        self.channel[data['ns']].member[data['user']]['con']-= 1
        if self.channel[data['ns']].member[data['user']]['con'] == 0:
            del self.channel[data['ns']].member[data['user']]
            
    def pkt_recv_kicked(self, data):
        # Get rid of users who get kicked.
        if not data['user'] in self.channel[data['ns']].member: return
        del self.channel[data['ns']].member[data['user']]
        
    def pkt_recv_privchg(self, data):
        # Got to make sure users' privclasses are up to date.
        if not data['user'] in self.channel[data['ns']].member: return
        self.channel[data['ns']].member[data['user']]['pc'] = data['pc']
    
    def pkt_disconnect(self, data):
        # dAmn also likes to disconnect clients. A lot.
        self.close()
        self.flag.connected = False
        
class ReconnectingClient(Client):
    """Persistant dAmn Client Class."""
    
    class flag(Client.flag):
        """Status flags are stored here"""
        autorejoin = True
        disconnecting = False
        reconnect = False
        quitting = False
        restart = False
        close = False
        debug = False
    
    autojoin = ['chat:Botdom']
    _disconnects = 0
    _connect_attempts = 0
    _connect_attempt_limit = 3
    
    session = None
    
    def __inst__(self):
        self.agent = 'dAmnViper (Python 3.1) ReconnectingClient/Client/dAmnSock/{0}.{1}'.format(
            self.platform.version, self.platform.build)
        self._received = time.strftime('%Y-%m-%d')
        self._connect_attempts = 0
        self.session = None
    
    def nullflags(self):
        Client.nullflags(self)
        self.flag.disconnecting = False
        self.flag.reconnect = False
        self.flag.quitting = False
        self.flag.restart = False
        self.flag.close = False
        
    def start(self, vals=None, *args, **kwargs):
        """Overwrite this method for a more customized client."""
        # We want this to be a continuous loop, so we can keep trying to connect if there are problems.
        while True:
            self._connect_attempts += 1
            if not self.authenticate(vals, *args, **kwargs):
                return
            self.logger('~Global', '** Opening a connection to dAmn.', False)
            if not self.make_connection(vals, *args, **kwargs):
                return
            # Enter the main loop of the program! Hells yeah!
            self.run(*args, **kwargs)
            # Ok, we're not connected anymore! What should we do?
            if self.attempt_reconnect(vals, *args, **kwargs):
                self.nullflags()
                continue
            break
    
    def run(self, *args, **kwargs):
        """Overwrite this method to provide your own application with a main loop."""
        if not self.sock:
            return
        while self.sock.connected():
            try:
                self.mainloop(*args, **kwargs)
                self.on_loop(*args, **kwargs)
            except KeyboardInterrupt as e:
                sys.stdout.write('\n')
                self.sock.close()
                self.flag.quitting = True
                self.process_packet('disconnect\ne=force quit (^C)\n\n', lambda data: data)
                break
        while len(self.sock.packet) > 0:
            self.mainloop(*args, **kwargs)
            self.on_loop(*args, **kwargs)
    
    def mainloop(self, *args, **kwargs):
        """Overwrite this if you want to do stuff during the main program loop."""
        self._received = time.strftime('%Y-%m-%d')
        self.get_packets()
        time.sleep(.05)
    
    def authenticate(self, vals=None, *args, **kwargs):
        if not self.user.token:
            # If we don't have an authtoken, try and grab one!
            self.logger('~Global', '** Retrieving authtoken, this may take a while...', False)
            self.get_token()
            if self.session is None:
                self.logger('~Global', '>> Insufficient login details provided.', False)
                return False
            if self.session.status[0] != 1:
                # Something went wrong! Maybe the user entered the wrong details?!
                self.logger('~Global', '>> Failed to get an authtoken.', False)
                self.logger('~Global', '>> {0}'.format(self.session.status[1]), False)
                return False
            # If we get here all is well.
            self.logger('~Global', '** Got an authtoken.', False)
        return True
    
    def attempt_reconnect(self, vals=None, *args, **kwargs):
        if self._connect_attempt_limit == 0:
            return True
        if self.flag.reconnect and self._connect_attempts < self._connect_attempt_limit:
            return True
        if self._connect_attempts >= self._connect_attempt_limit:
            n = self._connect_attempts
            self.logger('~Global', '>> Failed to connect to dAmn '+str(n)+' times. Giving up.', False)
            self.logger('~Global', '>> Make sure you entered the login details correctly.', False)
            self.logger('~Global', '>> If you already have, there may be a problem with dAmn.', False)
        return False
    
    def make_connection(self, vals=None, *args, **kwargs):
        # Actually try to connect to dAmn!
        # This simply opens a socket connection. The rest takes more effort.
        self.connect()
        if not self.flag.connecting:
            self.logger('~Global', '>> Failed to open a connection to dAmn.', False)
            return False
        # Send a handshake to the server!
        self.handshake(vals if bool(vals) else {'Client':'Viper.ReconnectingClient'})
        return True
        
    def pkt_login(self, data):
        Client.pkt_login(self, data)
        if data['e'] != 'ok':
            return
        self._connect_attempts = 0
        for ns in self.autojoin:
            self.join(self.format_ns(ns))
    
    def pkt_join(self, data):
        Client.pkt_join(self, data)
        if len(self.channel) > 0:
            return
        if data['e'] != 'ok':
            self.sock.append('disconnect\ne=no joined channels\n\n')
    
    def pkt_part(self, data):
        Client.pkt_part(self, data)
        if data['e'] != 'ok': return
        if self.channel or self.flag.disconnecting or self.flag.quitting:
            return
        self.sock.append('disconnect\ne=no joined channels\n\n')
    
    def pkt_kicked(self, data):
        del self.channel[data['ns']]
        if self.flag.disconnecting or self.flag.quitting:
            return
        if 'r' in data.keys():
            if 'autokicked' in data['r'].lower() or 'not privileged' in data['r'].lower():
                if len(self.channel) > 0:
                    return
                self.sock.append('disconnect\ne=no joined channels\n\n')
                return
        if self.flag.autorejoin:
            self.join(data['ns'])
    
    def pkt_disconnect(self, data):
        Client.pkt_disconnect(self, data)
        self._disconnects+= 1
        self.channel = {}
        if self.flag.quitting:
            self.flag.close = True
            return
        if data['e'] == 'no joined channels' and len(self.autojoin) == 0:
            return
        if self.flag.disconnecting:
            self.flag.disconnecting = False
        else:
            self.logger('~Global', '>> Experiencing an unexpected disconnect.', False)
            self.logger('~Global', '>> Attempting to reconnect in a moment.', False)
        time.sleep(1)
        self.nullflags()
        self.flag.reconnect = True
        
    def logger(self, ns, msg, showns=True, mute=False, pkt=None):
        stamp = time.strftime('%H:%M:%S')
        ns = ns + '|' if showns else ''
        if not mute or self.flag.debug:
            try:
                sys.stdout.write('{0}|{1} {2}\n'.format(stamp, ns, msg))
            except:
                sys.stdout.write('{0}| >> I failed to display a message! Sorry.\n'.format(stamp))
            sys.stdout.flush()

# EOF



''' Example of a very basic dAmn bot!
    Created by photofroggy
'''

from dAmnViper.base import ReconnectingClient

# Extend the dAmnViper.ReconnectingClient class to add some functionality.

class my_client(ReconnectingClient):
    
    def __inst__(self, username, password, admin, trigger='!', autojoin=None, callbacks=None):
        ReconnectingClient.__inst__(self) # Useful to call this.
        self.user.username = username
        self.user.password = password
        self._admin = admin.lower()
        self.trigger = trigger
        self.autojoin = autojoin if bool(autojoin) else ['Botdom']
        self.callbacks = callbacks if bool(callbacks) else {}
    
    def pkt_recv_msg(self, data):
        # Provide basic command firing.
        if data['message'][:len(self.trigger)] == self.trigger:
            data['message'] = data['message'][len(self.trigger):]
            data['args'] = data['message'].split(' ')
            self.callbacks.get(data['args'][0].lower(), self._no_cmd)(data, self)
    
    def _no_cmd(self, data, client):
        pass
    
def cmd_about(data, dAmn):
    """Basic command callback."""
    dAmn.say(data['ns'], data['user']+': Basic dAmn Viper bot by photofroggy.')

def cmd_quit(data, dAmn):
    """Quit command! Enter the admin name in place of 'admin'!"""
    if data['user'].lower() != dAmn._admin:
        return
    dAmn.say(data['ns'], data['user']+': Closing down!')
    dAmn.flag.quitting = True
    dAmn.disconnect()

def cmd_refresh(data, dAmn):
    """Quit command! Enter the admin name in place of 'admin'!"""
    if data['user'].lower() != dAmn._admin:
        return
    dAmn.say(data['ns'], data['user']+': Refreshing connection!')
    dAmn.flag.disconnecting = True
    dAmn.disconnect()
    
if __name__ == '__main__':
    # Create a client
    dAmn = my_client(
        'username', 'password', 'admin', '!',
        ['Botdom'], {'about': cmd_about, 'quit': cmd_quit, 'refresh': cmd_refresh}
    )
    # Start the dAmn client.
    dAmn.start()

# EOF



''' This is a basic example of how to create a basic dAmn client using dAmn Viper!
    Created by photofroggy
'''

from dAmnViper.base import Client

class client(Client):

    class user(Client.user):
        username = 'username'
        password = 'password'
    
    def __inst__(self):
        self.agent = 'Basic Client/1 (Python/3.0) dAmn Viper/' + self.platform.stamp
        self.logger('~Server', '** Retrieving authtoken, this may take a while...', False)
        self.get_token()
        if self.user.token == None:
            self.logger('~Server', '>> Failed to get authtoken.', False)
            return
        self.logger('~Server', '** Got an authtoken.', False)
        self.logger('~Server', '** Starting dAmn...', False)
        self.connect()
        if not self.flag.connecting:
            return
        self.handshake({'Client':'simple'})

    def mainloop(self):
        if not self.sock:
            return
        while self.sock.connected():
            datal = self.get_packets()
            if not datal: continue
            for data in datal:
                if not data: continue
                if hasattr(self, 'evt_' + data['event']):
                    getattr(self, 'evt_' + data['event'])(data['args'])
    
    # Some basic event handlers.
    def evt_login(self, data):
        if data['e'] == 'ok':
            self.join('chat:BotLab')
    
    def evt_join(self, data):
        if data['e'] == 'ok':
            self.say(data['ns'], 'Hello, world!')

# You would obviously need to make things more complicated to make it usable.

if __name__ == '__main__':
    app = client()
    app.mainloop()
    input('>> Press enter to exit...')
# EOF



''' dAmnViper.deviantART
    This module is part of the dAmnViper package.
    Created by photofroggy.
'''

import re
import urllib
import urllib2
import cookielib
    
class HTTPResponder:
    headers = {
        'Location': 'localhost',
        'Status': (0, False),
    }
    
    def __init__(self, url, data=''):
        self.url = url
        self.data = data
    
    def geturl(self, *args, **kwargs):
        return self.url

def fetch_cookie(obj, username, password, extras={'remember_me':'1'}, client='dAmnViper (python2.x) TokenGrabber/2'):
    extras.update({'username': username, 'password': password})
    obj.jar = cookielib.CookieJar()
    opener = urllib2.build_opener(urllib2.HTTPCookieProcessor(obj.jar))
    req = urllib2.Request(
        obj.url,
        urllib.urlencode(extras),
        {'User-Agent': client},
    )
    try:
        response = opener.open(req)
    except IOError as e:
        response = HTTPResponder('ConnectionError')
        response.headers['Status'] = (e.reason.errno, e.strerror)
    return response
    # Well, that was nice and easy :D

def fetch_channel(obj, url='http://chat.deviantart.com/chat/botdom', client='dAmnViper (python3.x) TokenGrabber/2'):
    opener = urllib2.build_opener(urllib2.HTTPCookieProcessor(obj.jar))
    req = urllib2.Request(
        url,
        urllib.urlencode({}),
        {'User-Agent': client})
    response = {}
    try:
        resp = opener.open(req)
        response = HTTPResponder(resp.geturl(), resp.read())
    except IOError as e:
        response = HTTPResponder('ConnectionError')
        response.headers['Status'] = (e.reason.errno, e.strerror)
    return response

# EOF



''' dAmn Viper - A Python API for dAmn.
    Copyright (C) 2009  Henry Rapley <froggywillneverdie@msn.com>
    Released under a GNU GPL.
'''

import os
import re

_pyver = float(os.sys.version[:3])
if _pyver >= 2.6 and _pyver < 3:
    from dAmnViper._auth26 import fetch_cookie
    from dAmnViper._auth26 import fetch_channel
elif _pyver >= 3:
    from dAmnViper._auth3 import fetch_cookie
    from dAmnViper._auth3 import fetch_channel
else:
    import sys
    sys.stdout.write('>>> Your Python install must be at least Python 2.6!\n')
    sys.stdout.write('>>> Python 3.1 is preferable.\n')
    sys.stdout.flush()
    sys.exit(1)


class Login:
    """This class uses given login data to fetch a deviantART cookie and authtoken."""
    
    url = 'https://www.deviantart.com/users/login'
    curl = 'http://chat.deviantart.com/chat/botdom'
    
    def __init__(self, username, password, extras={'remember_me':'1'}, client='dAmnViper (python 3.x) TokenGrabber/2'):
        """ Initialise the object. Fetch and process an authtoken. """
        self.jar = None
        self.cookie = None
        self.token = None
        self.response = None
        self.status = (0, 'Nothing has happened yet.')
        # Attempt to fetch the authoken!
        response = fetch_cookie(self, username, password, extras, client)
        # Process the response!
        if not self.valid_login_url(response):
            self.handle(response)
            return
        response = fetch_channel(self, self.curl, client)
        # Process the response!
        url = response.geturl()
        if url == 'ConnectionError' or url != self.curl:
            self.handle(response)
            return
        self.crop(username, response.data)
    
    def valid_login_url(self, response):
        url = response.geturl()
        if url == 'ConenctionError':
            return False
        if 'wrong-password' in url:
            return False
        if 'verify.deviantart.com' in url:
            return False
        return True
    
    def crop(self, username, data):
        match = re.search('"'+username+'", "([0-9a-f]{32})"', data, re.IGNORECASE)
        if match is None or match.group(1) is None:
            self.status = (2, 'Authtoken not given. Not sure why.')
            return
        self.token = match.group(1)
        self.status = (1, 'Authtoken retrieved!')
    
    def handle(self, response):
        """ Handle a login failure. """
        loc = response.geturl()
        print(loc)
        if 'verify.deviantart.com' in loc:
            self.status = (7,
                "This account has not yet been verified. To verify, please check the email you used to register the account (don't forget to search junk mail), and click the link the email deviantART sent you.")
            return
        if 'wrong-password' in loc:
            self.status = (4, 'Incorrect username or password provided.')
            return
        if loc == 'localhost':
            status = response.headers.get('Status')
            if status[0] == -2:
                self.status = (3, 'Could not connect to the internet.')
            else:
                self.status = (5, status[1])
            return
        self.status = (6, 'Something went wrong. I do not know why.')

# EOF



''' This is and even more basic dAmn client using dAmn Viper!
    Created by photofroggy
'''

from dAmnViper.base import Client

def main(user=None, password=None):
    if not user or not password:
        return
    client = Client()
    client.user.username = user
    client.user.password = password
    client.logger('~Server', '** Connecting...', False)
    client.get_token()
    if client.user.token == None:
        client.logger('~Server', '>> Login failed.', False)
        return
    client.connect()
    if not client.flag.connecting:
        return
    client.handshake({'Client':'simple'})
    
    # Main loop!
    while client.Socket.connected():
        packets = client.get_packets()
        if not packets:
            continue
        for data in packets:
            if data['event'] == 'login' and data['args']['e'] == 'ok':
                client.join('chat:Botlab')
            if data['event'] == 'join' and data['args']['e'] == 'ok':
                client.say(data['args']['ns'], 'Hello, world!')

if __name__ == '__main__':
    main('username', 'password')
    input('>> Press enter to exit...')
# EOF


''' Example of a client using dAmn Viper's ReconnectingClient class!
    Created by photofroggy
'''

from dAmnViper.base import ReconnectingClient

if __name__ == '__main__':
    dAmn = ReconnectingClient()
    dAmn.user.username = 'username'
    dAmn.user.password = 'password'
    dAmn.autojoin = ['Botdom']
    dAmn.start()

# EOF


''' Authtoken grabbing example.
    Created by photofroggy
    
    This is a basic example of how to get an authtoken using dAmn Viper!
    It also acts as an interactive authtoken grabber when run as a
    script, sort of.
'''

from dAmnViper.base import dAmnSock
from dAmnViper.deviantART import Login

def get_authtoken(un='username', pw='password'):
    clientstr = 'Authtoken Grabber/1 (Python/3.0) dAmn Viper/' + dAmnSock.platform.stamp
    session = Login(un, pw, client=clientstr)
    
    if session.token is None:
        print('>>',session.status[1])
    else:
        print('>> Token:',session.token)

if __name__ == '__main__':
    # Designed for use with Python 3.x
    # Change input to raw_input to make it work with earlier versions.
    un = input('>> Username: ').strip()
    pw = input('>> Password: ').strip()
    get_authtoken(un, pw)

# EOF



''' dAmnViper.stream module
    This module is part of the dAmnViper package.
    Created by photofroggy.
'''

import os
import time
import errno
import socket
from threading import Thread, BoundedSemaphore

class Connection(Thread):
    """ This is the Connection class. It handles the basic functionality
        of the connection between the bot and dAmn's server.
    """
    
    class server:
        """Server connection details."""
        host = None
        version = None
        port = None
    
    _charset = 'latin-1'
    _lock = None
    _runn = False
    _inn = False
    _debug = False

    def __init__(self, host='chat.deviantart.com', version='dAmnClient 0.3', port=3900, debug=False):
        """Initialise up in this motherfucka"""
        Thread.__init__(self)
        self.server.host = host
        self.server.version = version
        self.server.port = port
        self.sock = None
        self.packet = []
        self.__buffer = str()
        self._lock = BoundedSemaphore()
        self._debug = debug
    
    def connect(self):
        try:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.connect((socket.gethostbyname(self.server.host), self.server.port))
            self.sock.setblocking(0)
            self.sock.settimeout(.5)
        except socket.error:
            self.sock = None
            return
        self.start()
    
    def run(self):
        self._runn = True
        self._inn = True
        while self._runn:
            try:
                self.read()
            except Exception as e:
                self.close('disconnect\ne=python exception: {0}\n\n'.format(e.args[0]))
                self._runn = False
                break
            time.sleep(.05)
        self._inn = False
        
    def connected(self):
        # All this really does is check self.sock...
        open = bool(self.sock)
        if not open:
            self._runn = False
        return open
    
    def read(self):
        """Read packets from the socket stream, and store them in self.packet."""
        if not self.connected():
            self._runn = False
            return
        try:
            self._lock.acquire()
            data = self.sock.recv(8192)
            self._lock.release()
        except socket.error as e:
            self._lock.release()
            if e.args[0] == 'timed out':
                return
            if e.args[0] == errno.EAGAIN or (os.name == 'nt' and e.args[0] == errno.WSAEWOULDBLOCK):
                return
            if e.args[0] == errno.EINTR:
                return
            elif self.sock and e.args[0]:
                self.close('disconnect\ne=socket exception: {0}\n\n'.format(e.args[0]))
                return
        data = data.decode(self._charset, 'ignore')
        if not data and self.connected():
            self.close('disconnect\ne=socket closed\n\n')
            return
        self._lock.acquire()
        self.__buffer+= data
        data = self.__buffer.split("\0")
        self.__buffer = data.pop()
        self.packet.extend(data)
        pong = 'ping\n' in self.packet
        self._lock.release()
        if pong:
            self.send('pong\n')
            if self._debug:
                sys.stdout.write('>> Ping!')
    
    def close(self, packet=None):
        try: self.sock.close()
        except socket.error: pass
        self.sock = None
        if packet:
            self.packet.append(packet)
        self._runn = False
    
    def send(self, data):
        if not self.connected(): return
        try:
            self._lock.acquire()
            sent = self.sock.send((data+'\0').encode(self._charset, 'ignore'))
            self._lock.release()
            return sent
        except socket.error:
            return 0
    
    def get_packet(self):
        self._lock.acquire()
        val = False if len(self.packet) == 0 else self.packet.pop(0)
        self._lock.release()
        return val
    
    def get_packets(self):
        self._lock.acquire()
        L = len(self.packet)
        rets = [self.packet.pop(0) for i in range(0, L)]
        self._lock.release()
        return rets
    
    def append(self, pkt):
        self._lock.acquire()
        self.packet.append(pkt)
        self._lock.release()

# EOF



''' dAmnViper setup.
'''

from distutils.core import setup

setup(name='dAmnViper',
    version='2.54',
    description='Python dAmn bot',
    author='photofroggy',
    author_email='froggywillneverdie@msn.com',
    url='https://github.com/photofroggy/dAmnViper',
    packages=[
        'dAmnViper',
        'dAmnViper.examples'
    ],
    platforms=['Any'],
    classifiers=[
        'Natural Language :: English',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Topic :: Software Development :: Libraries :: Application Frameworks'
    ],
    long_description="""
==========
dAmn Viper
==========

**dAmn Viper** is a Python package created by photofroggy.

This package provides an API for connecting to and interacting with
the dAmn chats. That is, deviantART's messaging network chats.

The package was created using Python version 3, so should be used with
this version, but it has been tested and proven to work on Python 2.6 and
greater, thanks to some hacky module scripting.

-------------
Brief example
-------------

Creating a client quickly for dAmn is quite easy using dAmn Viper. Below
is one of the simplest examples of a client that simply connects, and
tries to stay connected::
    
    from dAmnViper.base import ReconnectingClient
    
    dAmn = ReconnectingClient()
    dAmn.user.username = 'username'
    dAmn.user.password = 'password'
    dAmn.autojoin = ['Botdom']
    dAmn.start()

That is all that is required! It is advised that you use the
ReconnectingClient class when making applications that connect to dAmn.
Chat bots and full clients can be made by extending the
ReconnectingClient class to add functionality, as shown in the examples
provided.

---------
Prospects
---------

I know it is a bad idea to not use Twisted, but this library was
intended for use in applications where the target audience is not likely
to want to spend time installing program after program. The idea was to
allow the user to get the application working as quickly as possible.

I will be experimenting with Twisted in the near future and may
re-release dAmn Viper using Twisted for managing the connection.

----------
DISCLAIMER
----------

Disclaimer::

		dAmn Viper is in no way affiliated with or endorsed by deviantART.com.
	This is not an official service of deviantART.com. This is an independent
	project created by Henry Rapley:
		<http://photofroggy.deviantart.com>
	
		THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
	APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
	HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
	OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
	THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
	PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
	IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
	ALL NECESSARY SERVICING, REPAIR OR CORRECTION.

"""
)

# EOF



''' dAmnViper.parse module
    This module is part of the dAmnViper package.
    Created by photofroggy.
'''

import re

class Packet:
    """ Use this class to parse dAmn packets.
        Data is stored in the attributes cmd, param,
        args, body and raw.
    """

    def __init__(self, data=None, sep='='):
        self.cmd, self.param, self.args, self.body, self.raw = None, None, {}, None, data
        if not bool(data): return
        if data.find('\n\n') != -1:
            self.body = data[data.find('\n\n')+2:]
            data = data[:data.find('\n\n')]
        breaks = data.split('\n')
        if not bool(breaks): return
        if len(breaks) >= 1 and not sep in breaks[0]:
            head = breaks.pop(0).split(' ')
            self.cmd = head[0] or None
            self.param = None if len(head) < 2 else head[1]
        for line in breaks:
            if line.find(sep) == -1: continue
            self.args[line[:line.find(sep)]] = line[line.find(sep)+len(sep):]
        # And that's the end of that chapter.

class Tablumps:
    """A very simple static class used to parse or capture dAmn-style Tablumps."""
    
    expressions = None
    replace = None
    titles = None
    subs = None
    
    def __init__(self):
        """Populate the expressions and replaces used when parsing tablumps."""
        if self.expressions is not None:
            return
        # Regular expression objects used to find any complicated tablumps.
        self.expressions = [
            re.compile("&avatar\t([a-zA-Z0-9-]+)\t([0-9]+)\t"),
            re.compile("&dev\t(.)\t([a-zA-Z0-9-]+)\t"),
            re.compile("&emote\t([^\t]+)\t([0-9]+)\t([0-9]+)\t(.*?)\t([a-z0-9./=-]+)\t"),
            re.compile("&a\t([^\t]+)\t([^\t]*)\t"),
            re.compile("&link\t([^\t]+)\t&\t"),
            re.compile("&link\t([^\t]+)\t([^\t]+)\t&\t"),
            re.compile("&acro\t([^\t]+)\t(.*)&\/acro\t"),
            re.compile("&abbr\t([^\t]+)\t(.*)&\/abbr\t"),
            re.compile("&thumb\t(?P<ID>[0-9]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t"),
            re.compile("&img\t([^\t]+)\t([^\t]*)\t([^\t]*)\t"),
            re.compile("&iframe\t([^\t]+)\t([0-9%]*)\t([0-9%]*)\t&\/iframe\t"),
        ]
        self.titles = ('avatar', 'dev', 'emote', 'a', 'link', 'link', 'acronym', 'abbr', 'thumb', 'img', 'iframe')
        # Regular expression objects used to find and replace complicated tablumps.
        self.subs = [
            (re.compile("&avatar\t([a-zA-Z0-9-]+)\t([0-9]+)\t"), ":icon\\1:"),
            (re.compile("&dev\t(.)\t([a-zA-Z0-9-]+)\t"), ":dev\\2:"),
            (re.compile("&emote\t([^\t]+)\t([0-9]+)\t([0-9]+)\t(.*?)\t([a-z0-9./=-]+)\t"), "\\1"),
            (re.compile("&a\t([^\t]+)\t([^\t]*)\t"), "<a href=\"\\1\" title=\"\\2\">"),
            (re.compile("&link\t([^\t]+)\t&\t"), "\\1"),
            (re.compile("&link\t([^\t]+)\t([^\t]+)\t&\t"), "\\1 (\\2)"),
            (re.compile("&acro\t([^\t]+)\t"), "<acronym title=\"\\1\">"),
            (re.compile("&abbr\t([^\t]+)\t"), "<abbr title=\"\\1\">"),
            (re.compile("&thumb\t([0-9]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t([^\t]+)\t"), ":thumb\\1:"),
            (re.compile("&img\t([^\t]+)\t([^\t]*)\t([^\t]*)\t"), "<img src=\"\\1\" alt=\"\\2\" title=\"\\3\" />"),
            (re.compile("&iframe\t([^\t]+)\t([0-9%]*)\t([0-9%]*)\t&\/iframe\t"), "<iframe src=\"\\1\" width=\"\\2\" height=\"\\3\" />"),
            (re.compile("<([^>]+) (width|height|title|alt)=\"\"([^>]*?)>"), "<\\1\\3>"),
        ]
        # Search and replace pairs used to parse simple HTML tags.
        self.replace = [
            ("&b\t", "<b>"),
            ("&/b\t", "</b>"),
            ("&i\t", "<i>"),
            ("&/i\t", "</i>"),
            ("&u\t", "<u>"),
            ("&/u\t", "</u>"),
            ("&s\t", "<s>"),
            ("&/s\t", "</s>"),
            ("&sup\t", "<sup>"),
            ("&/sup\t", "</sup>"),
            ("&sub\t", "<sub>"),
            ("&/sub\t", "</sub>"),
            ("&code\t", "<code>"),
            ("&/code\t", "</code>"),
            ("&p\t", "<p>"),
            ("&/p\t", "</p>"),
            ("&ul\t", "<ul>"),
            ("&/ul\t", "</ul>"),
            ("&ol\t", "<ol>"),
            ("&/ol\t", "</ol>"),
            ("&li\t", "<li>"),
            ("&/li\t", "</li>"),
            ("&bcode\t", "<bcode>"),
            ("&/bcode\t", "</bcode>"),
            ("&br\t", "\n"),
            ("&/a\t", "</a>"),
            ("&/acro\t", "</acronym>"),
            ("&/abbr\t", "</abbr>"),
        ]
    
    def parse(self, data):
        # Parse any dAmn Tablumps found in our input data.
        try:
            for lump, repl in self.replace:
                data = data.replace(lump, repl)
            for expression, repl in self.subs:
                data = expression.sub(repl, data)
        except Exception:
            pass
        return data
    
    def capture(self, text):
        # Return any dAmn Tablumps found in our input data.
        lumps = {}
        for key, expression in enumerate(self.expressions):
            cc = expression.findall(text)
            if not cc:
                continue
            lumps[self.titles[key]] = cc
        return lumps

class Protocol:
    """ Protocol processor.
        Use methods of this class to process dAmn information.
        You can customise how different packets are handled by
        adding itemds to protocol.maps and protocol.messages.
    """
    maps = {
        'dAmnServer': ['version'],
        'login': ['username', ['e'], 'data'],
        'join': ['ns', ['e'] ],
        'part': ['ns', ['e', '*r'] ],
        'property': ['ns', ['p', 'by', 'ts'], '*value' ],
        'recv_msg': [None, [('from', 'user')], '*message'],
        'recv_action': [None, [('from', 'user')], '*message'],
        'recv_join': ['user', None, '*info'],
        'recv_part': ['user', ['r']],
        'recv_privchg': ['user', ['by', 'pc']],
        'recv_kicked': ['user', ['by'], '*r'],
        'recv_admin_create': [None, ['p', ('by', 'user'), ('name', 'pc'), 'privs']],
        'recv_admin_update': [None, ['p', ('by', 'user'), ('name', 'pc'), 'privs']],
        'recv_admin_rename': [None, ['p', ('by', 'user'), 'prev', ('name', 'pc')]],
        'recv_admin_move': [None, ['p', ('by', 'user'), 'prev', ('name', 'pc'), ('n', 'affected')]],
        'recv_admin_remove': [None, ['p', ('by', 'user'), ('name', 'pc'), ('n', 'affected')]],
        'recv_admin_show': [None, ['p'], 'info'],
        'recv_admin_showverbose': [None, ['p'], 'info'],
        'recv_admin_privclass': [None, ['p', 'e'], 'command'],
        'kicked': ['ns', [('by', 'user')], '*r'],
        'ping': [],
        'disconnect': [None, ['e']],
        'send': ['ns', ['e']],
        'kick': ['ns', [('u', 'user'), 'e']],
        'get': ['ns', ['p', 'e']],
        'set': ['ns', ['p', 'e']],
        'kill': ['ns', ['e']],
        'unknown': [None, None, None, None, 'packet'],
    }
    names = [('recv', 'admin')]
    messages = {
        'dAmnServer': ('** Connected to dAmn Server {0}.', False),
        'login': ('** Login as {0}: {1}.', False),
        'join': ('** Join {ns}: "{1}".', False),
        'part': ('** Part {ns}: "{1}". [{2}]', False),
        'property': ('** Got {1} for {ns}.', False),
        'recv_msg': ('<{1}> {2}',),
        'recv_action': ('* {1} {2}',),
        'recv_join': ('** {1} has joined.',),
        'recv_part': ('** {1} has left. [{2}]',),
        'recv_privchg': ('** {1} has been made a member of {3} by {2} *',),
        'recv_kicked': ('** {1} has been kicked by {2} * {3}',),
        'recv_admin_create': ('** Privilege class {3} has been created by {2} with: {4}',),
        'recv_admin_update': ('** Privilege class {3} has been updated by {2} with: {4}',),
        'recv_admin_rename': ('** Privilege class {4} has been renamed to {3} by {2}',),
        'recv_admin_move': ('** All members of {3} have been moved to {4} by {2} -- {5} affected user(s)',),
        'recv_admin_remove': ('** Privilege class {3} has been removed by {2} -- {4} affected user(s)',),
        'recv_admin_show': None,
        'recv_admin_showverbose': None,
        'recv_admin_privclass': ('** Admin command "{2}" failed: {1}',),
        'kicked': ('** You have been kicked by {1} * {2}',),
        'ping': ('** Ping...', False, True),
        'disconnect': ('** You have been disconnected * {0}', False),
        'send': ('** Send error: {1}',),
        'kick': ('** Could not kick {1}: {2}',),
        'get': ('** Could not get {1}info for {ns}: {2}', False),
        'set': ('** Could not set {1}: {2}',),
        'kill': ('** Kill error: {1}',),
        'unknown': ('** Received unknown packet in {0}: {1}', False),
    }
    
    tablumps = Tablumps
    
    def __init__(self):
        self.tablumps = self.tablumps()
    
    def evt_namespace(self, pkt):
        # Work on the namespace of an event!
        namespace = pkt.cmd
        for conditions in self.names:
            if conditions[0] == namespace:
                subline = pkt.body.split('\n')[0]
                subline = subline.split(' ')
                if not subline or not subline[0]:
                    break
                namespace = '_'.join([namespace, subline[0]])
                if len(conditions) > 1 and len(subline) > 1:
                    if conditions[1] == subline[0]:
                        namespace = '_'.join([namespace, subline[1]])
                break
        return namespace if namespace in self.maps else 'unknown'
    
    def mapper(self, pkt):
        # Return pkt's data mapped according to the conditions defined in protocol.maps.
        store = {'args': {}, 'rules': [], 'event': self.evt_namespace(pkt)}
        map = self.maps[store['event']]
        return getattr(self, 'gen_'+pkt.cmd, self.sort)(store, pkt, map)
        
    def sort(self, store, data, map):
        # Sort data into store according to the conditions in map.
        for i, cond in enumerate(map):
            if not cond: continue
            if i is 0:
                ptab = cond[0] == '*'
                if ptab: cond = cond[1:]
                val = (data.param if not ptab else self.tablumps.parse(data.param)) if data.param else ''
                store['rules'].append((cond, val))
                store['args'][cond] = val
            if i is 1:
                for item in cond:
                    if not item:
                        continue
                    argn, name = item if isinstance(item, tuple) else (item, item)
                    ptab = argn[0] == '*'
                    if ptab and argn == name: name = argn[1:]
                    val = '' if not argn in data.args.keys() else data.args[argn]
                    if ptab: val = self.tablumps.parse(val)
                    store['rules'].append((name, val))
                    store['args'][name] = val
            if i is 2:
                ptab = cond[0] == '*'
                if ptab: cond = cond[1:]
                val = (data.body if not ptab else self.tablumps.parse(data.body)) if data.body else ''
                store['rules'].append((cond, val))
                store['args'][cond] = val
            if i is 3:
                store = self.sort(store, Packet(data.body), cond)
            if i is 4:
                val = data.raw if data.raw else ''
                store['rules'].append((cond, val))
                store['args'][cond] = val
        return store
    
    def logger(self, event, data, ns, pkt):
        # Return a log_list (channel, message[, bool(showns)[, bool(mute)]]).
        sequence = [ns, '', True, False, pkt]
        if not event in self.messages.keys():
            return None
        if hasattr(self, 'log_'+event):
            return getattr(self, 'log_'+event)(event, data, ns, pkt)
        msgtpl = self.messages[event]
        if msgtpl is None or len(msgtpl) == 0:
            return None
        data = [(item[1] if bool(item[1]) else '') for item in data]
        disp = (msgtpl[0].replace('{ns}', ns)).format(*data)
        if disp[-3:] == ' []':
            disp = disp[:-3]
        sequence[1] = disp
        msgtpl = [] if len(msgtpl) == 1 else msgtpl[1:]
        i = 2
        for item in msgtpl:
            sequence[i] = item
            i+=  1
        return sequence
    
    def gen_recv(self, store, data, map):
        # Generic recv packet mapping operations.
        store['args']['ns'] = data.param
        store['rules'].append(('ns', data.param))
        store = self.sort(store, Packet(data.body), map)
        if store['event'] in ('recv_msg', 'recv_action'):
            store['args']['raw'] = data.raw
            store['rules'].append(('raw', data.raw))
        return store

# EOF
